// author: Chuck Baker
// OPIE version 10.0

//var EXPORTED_SYMBOLS = [ "OPIE" ];

if (typeof(OPIE) == 'undefined') var OPIE = {};

//		Common (global) variables
var Cc = Components.classes;
var Ci = Components.interfaces;
var Cu = Components.utils;
var consoleLog = Cc["@mozilla.org/consoleservice;1"].getService(Ci.nsIConsoleService);

Cu.import("resource://gre/modules/AddonManager.jsm");
Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://gre/modules/osfile.jsm");

window.addEventListener("load",function(){OPIE.init()},true);
window.addEventListener("unload",function(){OPIE.unload()},true);

// Replacement for deprecated Application.storage
var OPIEstorage = {
	topWin: Cc["@mozilla.org/appshell/window-mediator;1"].getService(Ci.nsIWindowMediator).getMostRecentWindow("navigator:browser"),  
	setItem: function setItem(key, value){
			if(typeof this.topWin.storage == "undefined") this.topWin.storage = {};
			if(key == null || key == "") return false;
			this.topWin.storage[key] = value;
			return true;
	},
	getItem: function getItem(key,dFault){
			if(key == null || key == "") return false;
			if(typeof this.topWin.storage[key] == "undefined") return dFault;
			let value = this.topWin.storage[key];
			return value;
	}
}

OPIE = {
	opieProfDir: {}, 								// Pointer to profile directory
	opieExtNames: [],								// List of installed extensions
	opieRootEdits: [],							// List of edited preference roots
	OPIE_GUID: "OPIE@guid.customsoftwareconsult.com",
	opieMainWindow: {},							// Main browser window
	opieMsg: [],										// Opie messages
	opieImportList: [],							// List of prefs to import
	opieExportDir: new String(),		// Export directory path
	opieSelectionCount: 0,					// Number of items selected for export
	opieExtensionCount: 0,					// Number of extensions avaliable for export
	opieEditGUID: new String(),			// GUID of preference root being edited
	opieFileUtils: {},							// Pointer to "resource://gre/modules/FileUtils.jsm"
	opieNameGuidxref: [],						// Extension name/GUID cross reference array
	opieCellCoords: {},							// Row, column of a cell
	opieBoxObject: {},							// Pointer to the tree box object
	opieRootCells: [],							// Array of preference root cells
	opieStyles: [],									// Array of possible cell styles
	opieSysTmpDir: {},							// System tmp dir root
	opieTmpDir: {},									// Temp dir used to extract pref files
	opieWorkDir: {},								// Temp dir used to extract individual pref files
	opieOverwrite: Boolean(false),	// Overwrite existing exports?
	opieErrCnt: 0,									// Count of locking errors
	opieErr: 0,											// Error flag (0 - 4)
	opieExtDir: {},									// Pointer to the Fx profile extension directory
	opieWin: new Object(),

	opiePrefs: Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefService).getBranch(""),

	opieGetUnicharPref: function (prefName) {
		try{
			return OPIE.opiePrefs.getComplexValue(prefName,  Ci.nsISupportsString).data;
		}catch(err){
			alert(prefName+"\n\n"+err);
		}
		return true;
	},

	opieSetUnicharPref: function (prefName, prefValue) {
		let supportsString = Cc["@mozilla.org/supports-string;1"].createInstance(Ci.nsISupportsString);
		supportsString.data = prefValue;
		this.opiePrefs.setComplexValue(prefName, Ci.nsISupportsString, supportsString);
		return true;
	},

	opiePrefObj: function (Type,Name,Value){
		this.Type = Type;
		this.Name = Name;
		this.Value = Value;
	},

	opieAlertObj: function (Text,Style){
		this.Text = Text;
		this.Style = Style;
		return true;
	},

	init: function (){
		// Initialize variables and objects
		this.opieSetMsgs();
		
		// Get pointer to profile directory
		this.opieProfDir = Cc["@mozilla.org/file/directory_service;1"].getService(Ci.nsIProperties).get("ProfD", Ci.nsIFile);
					
		this.opieGetExtensionDataList();
		this.opieExportDir = this.opieGetUnicharPref("extensions.opie.exportDir");
		this.opieFileUtils = Cu.import("resource://gre/modules/FileUtils.jsm", {}).FileUtils;
		
		this.opieExtDir = Cc["@mozilla.org/file/directory_service;1"].getService(Ci.nsIProperties).get("ProfD", Ci.nsIFile);
		this.opieExtDir.append('extensions');

		//window.removeEventListener("load",OPIE.init,true);
		return true;
	},

	opieSetMsgs: function (){
		// Error messages, prompts, etc.
		let wm = Cc["@mozilla.org/appshell/window-mediator;1"].getService(Ci.nsIWindowMediator);  
		this.opieMainWindow = wm.getMostRecentWindow("navigator:browser"); 

		if(!this.opieMsg[0]){
			let bundle = Services.strings.createBundle("chrome://opie/locale/opie.properties");
			let enumerator = bundle.getSimpleEnumeration();
			while (enumerator.hasMoreElements()) {
				let string = enumerator.getNext().QueryInterface(Ci.nsIPropertyElement);
				let stringName = string.key;
				if(stringName.substr(0,3) != "msg") continue;
				let index = stringName.substr(3);
				index = parseInt(index);
				let value = string.value;
				this.opieMsg[index] = value;
			}
		}	

		this.opieStyles.push("normal");
		this.opieStyles.push("changed_valid");
		this.opieStyles.push("changed_invalid");
		this.opieStyles.push("unknown");
		return true;
	},

	opieErrWindow: function (){
		let prefName = "extensions.opie.errorResponse"; 
		this.opiePrefs.setIntPref(prefName,0);
		this.opieWin = window.open("chrome://opie/content/opieErr.xul", "Opie Error", "chrome,modal,dialog=no,alwaysRaised,centerscreen,resizable");	
		let resp = this.opiePrefs.getIntPref(prefName); 
		return resp;
	},

	opieSetErrorResp: function (resp){
		let prefName = "extensions.opie.errorResponse"; 
		this.opiePrefs.setIntPref(prefName,resp);
		return true;
	},

	opieSetErrorWindow: function (){
		this.opieSetMsgs();
		var d = document.getElementById("line1");
		let tmp = this.opieMsg[22].replace("%PREF%",OPIEstorage.getItem("opiePrefName","???")); 
		tmp = tmp.replace("%EXT%",OPIEstorage.getItem("opieExtensionName","???"));
		d.value = tmp;
		d = document.getElementById("line2"),
		d.value = this.opieMsg[23];
		d = document.getElementById("button1"),
		d.label = this.opieMsg[24];
		d = document.getElementById("button2"),
		d.label = this.opieMsg[25];
		d = document.getElementById("button3"),
		d.label = this.opieMsg[26];
		d = document.getElementById("button4"),
		d.label = this.opieMsg[27];
		return true;
	},

	opieStripSpaces: function (extName){
		// Remove spaces and undesirable characters from extension name
		let tmp = ""; 
		let mask = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789"; 
		mask += ".+-_";
		for (let i = 0; i <= extName.length;  i++){
			let c = extName.charAt(i); 
			if( mask.indexOf(c) != -1 ){tmp += c};
		}
		return tmp;
	},

	getOpieVersion: function () {
		AddonManager.getAddonByID(this.OPIE_GUID, function(addon) {
			OPIE.opieSetMsgs();
			document.getElementById("opieVersionID").value = OPIE.opieMsg[0] + ": " + addon.version;
		});
		return true;
	},

	opieSelFiles: function (){
		// Select files to import
		const nsIFilePicker = Ci.nsIFilePicker;
		let fp = Cc["@mozilla.org/filepicker;1"]
				 .createInstance(nsIFilePicker);
		fp.init(window, this.opieMsg[7], nsIFilePicker.modeOpenMultiple);
		let filter = "*.oprefs"; 
		fp.appendFilter(filter,filter);
		
		// Set default directory to last import directory if it is set, otherwise try the export directory
		let err = Boolean(false); 
		let dd = Cc["@mozilla.org/file/local;1"]
						 .createInstance(Ci.nsIFile);
		try{
			dd.initWithPath(this.opieGetUnicharPref("extensions.opie.importDir"));
		}catch(e){
			try{
				dd.initWithPath(this.opieGetUnicharPref("extensions.opie.exportDir"));
			}catch(e){err = true;}
		}
		if(!err){fp.displayDirectory = dd;}

		let rv = fp.show(); 
		this.opieImportList = [];
		let fnames = []; 
		if (rv == nsIFilePicker.returnOK){
			let files = fp.files; 
			while(files.hasMoreElements()) {
				var file = files.getNext()
					.QueryInterface(Ci.nsIFile);
				this.opieImportList.push(file.path);
				fnames.push(file.leafName);
			}
		}
		document.getElementById("opieImportFileID").value = fnames.join("\n");
		let importButton = document.getElementById("opieImportButton"); 
		if(fnames.length == 0){
			importButton.disabled = true;
		}else{
			importButton.disabled = false;
		}
		return true;
	},

	opieSelDir: function (){
		// Select export directory
		const nsIFilePicker = Ci.nsIFilePicker;
		let fp = Cc["@mozilla.org/filepicker;1"]
						.createInstance(nsIFilePicker);
		fp.init(window, this.opieMsg[4], nsIFilePicker.modeGetFolder);
		fp.appendFilters(nsIFilePicker.filterAll | nsIFilePicker.filterText);
		
		let extDir = ""; 

		// Set default directory to current export directory
		let prefName = "extensions.opie.exportDir"; 
		if(this.opiePrefs.prefHasUserValue(prefName)){
			extDir = this.opieGetUnicharPref(prefName);
			if(extDir != ""){
				let dd = Cc["@mozilla.org/file/local;1"]
								 .createInstance(Ci.nsIFile);
				let err = Boolean(false); 
				try {
					dd.initWithPath(extDir);
				}catch(e){err = true;}
				if(!err){fp.displayDirectory = dd;}
			}
		}
		let rv = fp.show(); 
		if (rv == nsIFilePicker.returnOK){
			rv = fp.file;
			extDir = rv.path;
			this.opieSetUnicharPref("extensions.opie.exportDir", extDir);
			document.getElementById("opieExportDirID").setAttribute("value",extDir);;
		}
		return true;
	},

	opiePlaySound: function (which){
		if(!this.opiePrefs.getBoolPref("extensions.opie.playSounds")) return true;
		let soundURL = "chrome://opie/content/sounds/"+which+".ogg"; 
		let audio = new Audio(soundURL); 
		audio.play();
		return true;
	},

	opieRemoveStyle: function (txt){
		// remove ,style> tag from txt
		let styleStart = txt.indexOf("<style>"); 
		if(styleStart == -1){return txt;}
		let styleEnd = txt.indexOf("</style>"); 
		if(styleEnd == -1){return txt;}
		return txt.substr(styleEnd+8);
	},

	logMsg: function(msg) {
		msg = "OPIE: "+msg;
		console.log(msg)
		let cs = Cc["@mozilla.org/consoleservice;1"].getService(Ci.nsIConsoleService);
		cs.logStringMessage(msg);
		return true;
	},
	
	logMsg2: function(msg) {
		msg = "OPIE: "+msg;
		console.log(msg)
		let cs = Cc["@mozilla.org/consoleservice;1"].getService(Ci.nsIConsoleService);
		cs.logMessage(msg);
		return true;
	},

		
	propsConsole: function(obj,title,ignoreProperties,ignoreFunctions,ignoreObjects,expandObjs,sort){
		if(title != null){
			title = "("+title+")";
		}else{
			title = "";
		}
		if (obj === null) {
			this.logMsg("propsConsole called with null argument: "+title+"\n");
			return false;
		}
		if (obj === undefined) {
			this.logMsg("propsConsole called with undefined argument: "+title+"\n");
			return false;
		}
		let tmp, line, pVal, type, msg, attrb = [], func = [], objects = [];
		msg = "propsConsole: "+title+"\ntypeof '"+obj+"': ("+typeof obj+"), "+((sort) ? "Sorted" : "Unsorted")+"\n";
		for(let property in obj){
			line = property;
			try{
				pVal = obj[property];
			}catch(err){
				pVal = "(Unreachable)  "+err;
			}
			try{
				type = typeof obj[property];
			}catch(e){
				type = "Unknown";
			}
			try{
				tmp = line + " = "+pVal;
			}catch(e){
				tmp = "(Unreachable)";
			}
			switch(type){
				case 'function':
					if(!ignoreFunctions) func.push(tmp);
					break;
				case 'object':
					if(ignoreObjects) continue;
					try{
						let props = JSON.stringify(obj[property]);
						if(expandObjs && props != "null") tmp += " --> "+props;
					}catch(err){;;}
					objects.push(tmp);
					break;
				default:
					if(!ignoreProperties) attrb.push(tmp);
					break;
			}
		} 
		if(func.length == 0) func.push("(none)");
		if(objects.length == 0) objects.push("(none)");
		if(attrb.length == 0) attrb.push("(none)");

		if(sort){
			attrb.sort();
			func.sort();
			objects.sort();
		}

		if(!ignoreProperties){
			msg += "\n*** PROPERTIES/ATTRIBUTES ***\n";
			for(let i in attrb){
				msg += attrb[i]+"\n";
			}
		}
		
		if(!ignoreFunctions){
			msg += "\n*** FUNCTIONS ***\n";
			for(let i in func){
				msg += func[i]+"\n";
			}
		}
		
		if(!ignoreObjects){
			msg += "\n*** OBJECTS ***\n";
			for(let i in objects){
				msg += objects[i]+"\n";
			}
		}
		this.logMsg(msg);
		return true;
	},

	// Main
	opieCellCoordsObj: function (row, col){
		this.row = row;
		this.col = col;
	},

	opieOpen: function (){
		window.openDialog('chrome://opie/content/opie.xul', 'OPIE', 'chrome,titlebar,toolbar,centerscreen,resizable');	
	},

	opieExportPane: function (){
		document.getElementById("opiebutton2ID").disabled = false;
		document.getElementById("opiebutton3ID").disabled = false;
		document.getElementById("opiebutton4ID").disabled = false;
		this.opieClearExportList();
		this.opieBuildExportList();
	},

	opieImport: function (){
		let opieResults = []; 
		let n = 0; 
		if(this.opieImportList == ""){
			let style =  "<style>color: red;font-size: 10pt;</style>";
			let tmp = style+document.getElementById("opie.import.file.default").value; 
			this.opieAlert(tmp);	// No extensions selected
			return false;
		}
		for(var i in this.opieImportList){
			let prefFile = this.opieImportList[i]; 
			n++;

			// Read the data
			let dataFile = Cc["@mozilla.org/file/local;1"]
											 .createInstance(Ci.nsIFile);
			dataFile.initWithPath(prefFile);
			
			// open an input stream from file
			let istream = Cc["@mozilla.org/network/file-input-stream;1"]
										.createInstance(Ci.nsIFileInputStream);
			istream.init(dataFile, 0x01, 0444, 0);
			istream.QueryInterface(Ci.nsILineInputStream);
		
			let cis = Cc["@mozilla.org/intl/converter-input-stream;1"]
								.createInstance(Ci.nsIConverterInputStream);
		
			let charset = "UTF-8"; 
			cis.init(istream,charset, 1024, Ci.nsIConverterInputStream.DEFAULT_REPLACEMENT_CHARACTER);
			let lis = cis.QueryInterface(Ci.nsIUnicharLineInputStream); 

			// read lines into array and parse into individual fields
			let line = {}, hasmore, pcnt = 0; 
			do {
				hasmore = lis.readLine(line);
				let tmp = line.value; 
				if(tmp.length == 0){break;}
				
				// Write the preferences
				let pref = JSON.parse(tmp);   
				switch(pref.Type){
					case "boolean":
						this.opiePrefs.setBoolPref(pref.Name,Boolean(pref.Value)); 
						break;
					case "integer":
						this.opiePrefs.setIntPref(pref.Name,parseInt(pref.Value));
						break;
					case "string":
						this.opieSetUnicharPref(pref.Name, String(pref.Value));
						break;
				}
				pcnt++;
			} while(hasmore);
			
			opieResults.push(dataFile.leafName+":\t"+pcnt+" "+this.opieMsg[8])

			istream.close();
			cis.close();	
		}
		
		opieResults.push(" ");
		opieResults.push(this.opieMsg[9].replace(/%NUM%/,n));
		document.getElementById("opieImportsResultsID").value = opieResults.join("\n");
		this.opiePlaySound('success');
		return true;
	},

	opieSelect: function (bool){
		// Select/deselect all
		for(let i in this.opieExtNames){
			let item = document.getElementById(i); 
			if(item){item.setAttribute("checked",bool);}
		}
		if(bool){
			this.opieSelectionCount = this.opieExtensionCount;
		}else{
			this.opieSelectionCount = 0;
		}
		this.setExportCaption();
	},

	opieGetExtensionDataList: function (){
		// Get a list of all installed extensions that have preferences (as deteermined by the existance of defaults/preferences/*.js
		this.opieExtNames = [];
		AddonManager.getAddonsByTypes(["extension"],function(list) {
			list.forEach(function(addons) {
				AddonManager.getAddonByID(addons.id, function(addon) {
					if(addon.optionsURL){	// If true, assume preferences exist
						let datafileName = OPIE.opieStripSpaces(addons.name)+"{" + addons.version + "}" + ".oprefs";
						let item = {};
						item.Guid = addons.id;
						item.Name = addons.name;
						item.DatafileName = datafileName;
						item.Version = addons.version;
						item.isXpi = OPIE.setIsXpi(item);
						item.hasPrefs = OPIE.hasPrefs(addon,item.isXpi);
						if(item.hasPrefs){
							OPIE.opieExtNames[addons.id] = item;
							OPIE.opieNameGuidxref[addons.name] = addons.id; 
						}
					}
				});
			});
		});
		return true;
	},
	
	setIsXpi: function(item){
		// If the extension location is a directory, it is not an .xpi file
		let id = item.Guid;
		let curExtDir = this.opieExtDir.clone();
		curExtDir.append(id);
		let isXpi;
		if(curExtDir.exists() && curExtDir.isDirectory()){
			isXpi = false;
		}else{
			isXpi = true;
		}
		return isXpi;		
	},
	
	hasPrefs: function (item,isXpi){
		let curExtDir = OPIE.opieExtDir.clone();
		let ok = false, entries;
		if(isXpi){
			curExtDir.append(item.id+".xpi");
			let xpiFile = curExtDir;

			let zipR = Cc["@mozilla.org/libjar/zip-reader;1"].createInstance(Ci.nsIZipReader);
			zipR.open(xpiFile);

			entries = zipR.findEntries("defaults/preferences/*.js$");
			ok = entries.hasMore();
		}else{
			curExtDir.append(item.id);
			curExtDir.append('defaults');
			curExtDir.append('preferences');
			if(curExtDir.exists()){
				entries = curExtDir.directoryEntries;
				ok = entries.hasMoreElements();
			}
		}
		return ok;
	},
	
	opieExportPrefs: function (){
		// Get and verify export directory
		if(!this.opieGetExportDir()){return false;}
		this.createTmpDir();
		// Get 'combined' file name
		let type = this.opiePrefs.getCharPref('extensions.opie.exportOption'); 
		let combined = (this.opiePrefs.getCharPref('extensions.opie.exportOption') == "combined") ? true : false; ;
		let name = document.getElementById("opieExportCombinedFilenameID").value; 
		if(combined){
			if(name.length == 0){
				let style = "<style>color: red;font-size: 10pt;</style>";
				let tmp = style+this.opieMsg[16]; 
				this.opieAlert(tmp);	// No name for combined export
				return false;
			}		
		}
		
		// Get overwrite option
		this.opieOverwrite = this.opiePrefs.getBoolPref('extensions.opie.overwrite');
		
		// Build an array of the individual preferences
		let numSelected = 0; 
		let combinedList = []; 
		let opieExtCount = 0; 
		let tmp = "";
		for(let i in this.opieExtNames){
			let listItem = document.getElementById(i); 
			let isChecked = (listItem.getAttribute("checked") == "true") ? true : false; 
			if(!isChecked) continue;
			let item = this.opieExtNames[i]; 
			let prefs = this.readPrefFiles(item);	// An array of preference names
			if(prefs === false) continue;
			if(prefs.length == 0){
				tmp = this.opieMsg[12].replace("%EXT%",item.Name);
				this.opieAlert(tmp,true);
				continue;
			}
			numSelected++;
			if(combined){
				combinedList = combinedList.concat(prefs);
				opieExtCount++;
			}else{
				if(this.opieExport(this.opieExportDir.path,item,prefs)){opieExtCount++;}
			}
		}

		if(numSelected == 0){
			let style = "<style>color: red;font-size: 10pt;</style>";
			tmp = style+this.opieMsg[02]; 
			this.opieAlert(tmp);	// No extensions selected
			return false;
		}
		
		let style = "<style>color: blue;font-size: 10pt;</style>";
		if(combined){
			name += ".oprefs";
			let ok = this.opieExportCombined(this.opieExportDir.path,name,combinedList); 
			if(ok){
				tmp = this.opieMsg[17].replace(/%NUM%/,opieExtCount); 
				tmp = style+tmp.replace(/%NAME%/,name);
			}else{
				tmp = this.opieMsg[15].replace(/%NAME%/,name); 
				let style = "<style>color: red;font-size: 10pt;</style>";
				tmp = style+tmp;
			}
		}else{
			tmp = style+this.opieMsg[06].replace(/%NUM%/,opieExtCount); 
		}
		if(this.opieErr != 4){this.opieAlert(tmp);}
		this.opieClearExportList();
		this.opieBuildExportList();
		this.clearTmpDir();
		this.opieErrCnt = 0;
		return true;
	},
	
	readyTmpDir: function(id){
		// Create a working directory named the same as the add-ons guid 
		this.opieWorkDir = this.opieTmpDir.clone();
		this.opieWorkDir.append(id);
		try{
			this.opieWorkDir.createUnique(Ci.nsIFile.DIRECTORY_TYPE, this.opieFileUtils.PERMS_DIRECTORY);
		}catch(err){
			let style =  "<style>color: red;font-size: 10pt;</style>";
			let tmp = style+this.opieMsg[18]; 
			tmp += "\n"+this.opieWorkDir.path+"\n";
			tmp += this.opieMsg[19]+"\n\n"; 
			tmp += err.message;
			this.opieAlert(tmp,true);
			return false;
		}
		return true;
	},
	
	createTmpDir: function(){
		// Create a tmp directory in the system temp folder
		this.opieSysTmpDir = this.opieFileUtils.getDir("TmpD", [], true);
		this.opieTmpDir = this.opieSysTmpDir.clone();
		this.opieTmpDir.append("opieTmp");
		this.opieTmpDir.createUnique(Ci.nsIFile.DIRECTORY_TYPE, this.opieFileUtils.PERMS_DIRECTORY);
		this.opieTmpDir.permissions = 511;	// Try to set permissions to drwxrwxrwx
		return true;
	},

	clearTmpDir: function(){	
		this.opieTmpDir.remove(true);
		return true;
	},
	
	unload: function(){
		// See if any previous tmp files exist and delete them
		this.opieSysTmpDir = this.opieFileUtils.getDir("TmpD", [], true);
		let entries = this.opieSysTmpDir.directoryEntries;
		while(entries.hasMoreElements()){
			let entry = entries.getNext();
			entry.QueryInterface(Ci.nsIFile);
			let file = Cc["@mozilla.org/file/local;1"].createInstance(Ci.nsIFile);
			file.initWithPath(entry.path);
			if(file.leafName.substr(0,7) != "opieTmp") continue;
			file.remove(true);
		}
		return true;
	},
	
	readPrefFiles: function(item){
		// Get the preference file from 'defaults/preferences/'.  Extract the file to a tmp dir if the extension is an xpi
		let entry, file, entries, zipR, prefFile, outFile, prefs = [], curExtDir, xpiFile, tmpDir, id = item.Guid;
		if(item.isXpi){
			// Extract all the prefs files into a tmp directory
			if(!this.readyTmpDir(id)){
				return false;
			}else{
				tmpDir = this.opieWorkDir;
			}
			curExtDir = this.opieExtDir.clone();
			curExtDir.append(id+".xpi");
			xpiFile = curExtDir;
			
			// Unzip the xpi's preference files
			zipR = Cc["@mozilla.org/libjar/zip-reader;1"].createInstance(Ci.nsIZipReader);
			zipR.open(xpiFile);

			entries = zipR.findEntries("defaults/preferences/*.js$");
			while (entries.hasMore()) {
				entry = entries.getNext();
				prefFile = entry.substr(entry.lastIndexOf("/")+1);
				outFile = tmpDir.clone();
				outFile.append(prefFile);
				zipR.extract(entry,outFile);
				outFile.permissions = 511;	// Try to set permissions to drwxrwxrwx
			}
			zipR.close();		
			curExtDir	= tmpDir.clone();
		}else{
			curExtDir = this.opieExtDir.clone();
			curExtDir.append(id);
			curExtDir.append('defaults');
			curExtDir.append('preferences');
		}
		if(!curExtDir.exists()) return false;
		
		// Parse the dir for all pref files and extract the prefs
		entries = curExtDir.directoryEntries;
		while(entries.hasMoreElements()){
			entry = entries.getNext();
			entry.QueryInterface(Ci.nsIFile);
			file = Cc["@mozilla.org/file/local;1"].createInstance(Ci.nsIFile);
			file.initWithPath(entry.path);
			let fis = Cc["@mozilla.org/network/file-input-stream;1"].createInstance(Ci.nsIFileInputStream);
			fis.init(file, 0x01, 0x124, 0);
			fis.QueryInterface(Ci.nsILineInputStream);

			let cis = Cc["@mozilla.org/intl/converter-input-stream;1"].createInstance(Ci.nsIConverterInputStream);
			cis.init(fis,"UTF-8", 0, 0x0000);
			
			let lis = cis.QueryInterface(Ci.nsIUnicharLineInputStream);
			let line = {}, hasmore;	
			do {
				hasmore = lis.readLine(line);			
				let prefLine = line.value.replace(/'/g,'"');	// Replace single quotes to double quotes
				if(prefLine.substr(0,6) != 'pref("') continue;
				let match, pref;
				pref = prefLine.substr(6);
				let quotePos = pref.indexOf('"');
				pref = pref.substr(0,quotePos);
				prefs.push(pref);
			} while(hasmore);
			cis.close();
			fis.close();
		}
		return prefs;
	},

	opieClearExportList: function (){
		let theTree = document.getElementById("opieExportList"); 
		let children = theTree.childNodes; 
		let n = children.length; 
		
		// Clear existing items
		for (let i=0; i < n;  i++) {
			theTree.removeChild(children[0]);
		}
		return true;
	},

	opieBuildExportList: function (){
		let theList = document.getElementById("opieExportList");
		this.opieSelectionCount = 0;
		this.opieExtensionCount = 0;
		
		
		// Sort the tree
		let sortArray = []; 
		for(var i in this.opieExtNames){
			sortArray.push(this.opieExtNames[i].Name);
		}
		sortArray.sort();
		
		for(let i in sortArray){
			let data = this.opieExtNames[this.opieNameGuidxref[sortArray[i]]]; 
			if(!data.hasPrefs) continue;	// Ignore extensions with no prefs
			
			let row = document.createElement('listitem'); 
			row.setAttribute('id', data.Guid );
			row.setAttribute('type', "checkbox" );
			row.setAttribute('label', data.Name+" (v"+data.Version+")");
			row.setAttribute('checked', "false" );
			row.onclick = function(event){OPIE.selectCount(this.checked)};
			theList.appendChild( row );
			this.opieExtensionCount++;
		}
		this.setExportCaption();
		if(this.opieExtensionCount == 0) this.opieAlert(this.opieMsg[03]);
		return true;
	},
	
	selectCount: function(checked){
		if(checked){
			this.opieSelectionCount++;
		}else{
			this.opieSelectionCount--;
		}
		this.setExportCaption();
	},
	
	setExportCaption: function(){
		let d = document.getElementById('extcount');
		let value = this.opieMsg[20].replace(/%CNT%/,this.opieExtensionCount);
		value = value.replace(/%SEL%/,this.opieSelectionCount);
		d.value = value;
		return true;
	},

	opieExport: function (dir,which,prefList){
//this.logMsg("opieExport: "+dir+","+JSON.stringify(which)+","+prefList)
//Cc["@mozilla.org/consoleservice;1"].getService(Ci.nsIConsoleService).logMessage(dir,which,prefList)
//console.log(dir,which,prefList)
//this.logMsg(JSON.stringify(dir))

		// Build export data from 'prefList' and write it to 'which.DatafileName' located in 'dir'
		let opieDataFile = Cc["@mozilla.org/file/local;1"].createInstance(Ci.nsIFile);
		opieDataFile.initWithPath(dir);
		opieDataFile.append(which.DatafileName);
		if(!this.opieOverwrite && opieDataFile.exists()){return false;}

		let dataFile = Cc["@mozilla.org/network/file-output-stream;1"].createInstance(Ci.nsIFileOutputStream);
		dataFile.init(opieDataFile, 0x02 | 0x08 | 0x20, 0755, 0); // write, create, truncate
		
		let charset = "UTF-8"; 
		let os = Cc["@mozilla.org/intl/converter-output-stream;1"]
			 .createInstance(Ci.nsIConverterOutputStream);

		os.init(dataFile, charset, 4096, 0x0000);
		let list = this.opieBuildExportFileData(prefList); 
		if(list != ""){os.writeString(list);}	
		os.close();
		if(this.opieErr == 4){return false;}
		return true;
	},

	opieExportCombined: function (dir,fname,prefList){
		let opieDataFile = Cc["@mozilla.org/file/local;1"]
			.createInstance(Ci.nsIFile);
		opieDataFile.initWithPath(dir);
		opieDataFile.append(fname);
		if(!this.opieOverwrite && opieDataFile.exists()){return false;}

		let dataFile = Cc["@mozilla.org/network/file-output-stream;1"]
			.createInstance(Ci.nsIFileOutputStream);
		dataFile.init(opieDataFile, 0x02 | 0x08 | 0x20, 0755, 0); // write, create, truncate
		
		let charset = "UTF-8"; 
		let os = Cc["@mozilla.org/intl/converter-output-stream;1"]
			 .createInstance(Ci.nsIConverterOutputStream);

		os.init(dataFile, charset, 4096, 0x0000);
		let list = this.opieBuildExportFileData(prefList); 
		if(list != ""){os.writeString(list);}	

		os.close();
		return true;
	},

	opieGetExportDir: function (){
		// Get and verify export directory
		var dir = document.getElementById("opieExportDirID").value
		if(dir == ""){
			let style = "<style>color: red;font-size: 10pt;</style>";
			let tmp = style+this.opieMsg[1]; 
			this.opieAlert(tmp); // No install directory	
			return false;
		}
		
		// Verify export directory
		let file = Cc["@mozilla.org/file/local;1"].createInstance(Ci.nsIFile);                 
			let err = new Boolean(false); 
		try {
			file.initWithPath(dir);
		}catch(e){err = true;}

		if (!(file.exists() && file.isDirectory()) || !err){
			let style = "<style>color: red;font-size: 10pt;</style>";
			let tmp = style+this.opieMsg[05].replace(/%DIR%/,dir); 
			this.opieAlert(tmp);
			return false;
		}
		this.opieExportDir = file;
		return true;
	},

	opieBuildExportFileData: function (prefList){
		// Return export data built from 'prefList'
		let prefService = Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefService); 
		let prefBranch = prefService.getBranch(null); 		
		let fileData = "";
		for(let i=0; i<prefList.length; i++){
			let pref = prefList[i];
			let type = prefBranch.getPrefType(pref);
			let value = this.opieGetExtPref(pref,type);
			
			switch(this.opieErr){
				case 0:	// No error
					break;
				case 1:	// Skip this preference and continue exporting
					continue;
					break;
				case 2:	// Skip this extension and continue exporting
					return "";
					break;
				case 3:	// Always ignore this error (can be changed in OPIE options)
					this.opiePrefs.setBoolPref("extensions.opie.ignoreError", true);
					continue;
					break;
				case 4:	// Abort this export session
					var style = "<style>color: red;font-size: 10pt; font-weight: bold;</style>";
					let tmp = style+this.opieMsg[28]+"\n"; 
					style = "<style>color: black;font-size: 10pt; font-weight: normal;</style>";
					tmp += style+this.opieMsg[29];
					this.opieAlert(tmp);
					return "";
					break;
			}

			let Type;
			let prefs = Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefBranch);
			switch (type){
				case prefs.PREF_BOOL:	
					Type = "boolean";
					break;
				case prefs.PREF_INT:
					Type = "integer";
					break;
				case prefs.PREF_STRING:
					Type = "string";
					break;
			}
			let line = new this.opiePrefObj(Type,pref,value); 
			fileData += JSON.stringify(line) + "\n";
		}
		return fileData;
	},
	
	opieGetExtPref: function (prefName,type){
		// Get the extension preference
		this.opieErr = 0;
		let opieErrCnt = 0;
		try{
			let prefValue;
			switch(type){
				case 32:	// String
					prefValue = this.opieGetUnicharPref(prefName);
					break;
				case 64:	// Integer
					prefValue = this.opiePrefs.getIntPref(prefName);
					break;
				case 128:	// Boolean
					prefValue = this.opiePrefs.getBoolPref(prefName);
					break;
			}
			return prefValue;
		}catch(e){
			opieErrCnt++;
			let ignoreError = this.opiePrefs.getBoolPref("extensions.opie.ignoreError"); 
			let err = e.toString().indexOf("NS_ERROR_UNEXPECTED"); 
			if(ignoreError && err != -1){
				this.opieErr = 1;
				return "";
			}else{
				if(err == -1){
					alert(e);	// Non-locking error
					this.opieErr = 4;
					return "";
				}else{
					this.opieErr = this.opieErrWindow();
				}
			}
		}
	},

	opieViewExDir: function (){
		let folder = document.getElementById("opieExportDirID").value;
		let nsLocalFile = Components.Constructor("@mozilla.org/file/local;1","nsILocalFile", "initWithPath");
		new nsLocalFile(folder).reveal();
		return true;
	},
	
	buttonCheck: function(){
		let folder = document.getElementById("opieExportDirID").value;
		let button1 = document.getElementById("opieviewbudirbtnID");
		let button2 = document.getElementById("opieclearbudirbtnID");
		let disabled = false;
		if(!folder) disabled = true;
		button1.disabled = disabled;
		button2.disabled = disabled;
		return true;
	},

	opieClearExDir: function (){
		if(this.opiePrefs.getBoolPref('extensions.opie.warnbeforeclear')){
			this.opieConfirmClear();
			if(!this.opiePrefs.getBoolPref('extensions.opie.confirm')) return false;
		}
		
		if(!this.opieGetExportDir()){return false;}

		let entries = this.opieExportDir.directoryEntries; 
		while(entries.hasMoreElements()){
			let entry = entries.getNext(); 
			entry.QueryInterface(Ci.nsIFile);
			if(entry.isFile()){entry.remove(false);}
		}
		let style = "<style>color: blue;font-size: 10pt;</style>";
		let tmp = style+this.opieMsg[14]; 
		this.opieAlert(tmp);
		return true;
	},
	
	opieHelp: function (helpIndex,fieldName){
		let opts = this.opieSetWinOpts(false);
		this.opieWin = window.openDialog('chrome://opie/content/opieHelp.xul', 'OPIE Help', opts,helpIndex,fieldName);
		this.opieWin.focus();
		return true;
	},

	opieAlert: function (message,modal){
		let opts = this.opieSetWinOpts(modal);
		this.opieWin = window.openDialog('chrome://opie/content/opieAlert.xul','OPIE Alert',opts,message);	
		this.opieWin.focus();
		return true;
	},

	opieConfirmClear: function (){
		let opts = this.opieSetWinOpts(true);
		this.opieWin = window.openDialog('chrome://opie/content/opieConfirmClear.xul','OPIE Confirm',opts);	
		this.opieWin.focus();
		return true;
	},
	
	setConfirmValue: function (bool){
		this.opiePrefs.setBoolPref('extensions.opie.confirm',bool);
	},
	
	setConfirmMsg: function (){
		this.opieSetMsgs();
		let msgBox = document.getElementById("opieMsgBox");
		let labelElm = document.createElement('label');
		let label = labelElm.cloneNode(false);
		label.value = this.opieMsg[29];
		label.className = 'fontwarn';
		msgBox.appendChild(label);
		label = labelElm.cloneNode(false);
		label.value = this.opieMsg[30];
		label.className = 'fontnormal';
		msgBox.appendChild(label);
		OPIE.opiePlaySound('beep');
		return true;
	},

	opieSetWinOpts: function(modal){
		let opts = "chrome,dialog,alwaysRaised,resizable";
		if(modal){opts += ",modal";}
		let opieMsgType = this.opiePrefs.getCharPref('extensions.opie.alertType');
		if(opieMsgType == 'slider'){opts += ",popup";}
		if(opieMsgType == 'standard'){opts += ",centerscreen,titlebar";}
		return opts;
	},

	opieSetAlertMsg: function (){
		// If the message contains a style or class, it must be the first part of the message 
		// (one style per line, if needed).  
		// If a class is assigned, it must be in the form of: <class>classname</class>
		let tmp = window.arguments[0]+"   \n\n";
		let lines = tmp.split(/\n/); // Break it into an array
		let msg = [], re, a;
		let separator = document.createElement('separator');
		for(let i=0;i<lines.length;i++){	
			let line = lines[i];
			let item = {};
			item.Style = "font-family: arial; font-size: 10pt; font-weight: normal; color: purple";
			item.Text = line;

			// Parse for links
			let link = line.indexOf("<link>");
			if(link == 0){
				re = /(<link>)(.*),(.*)(<\/link>)/g
				a = re.exec(line);
				item.Text = a[2];
				item.Link = a[3];
				item.Klass = "link";	
				msg[i] = item;
				continue;
			}
			
			// Parse for styles
			let style = line.indexOf("<style>");
			if(style == 0){
				re = /(<style>)(.*)(<\/style>)(.*)/g
				a = re.exec(line);
				item.Style = a[2];
				item.Text = a[4];
				msg[i] = item;
				continue;
			}
			
			// Parse for class
			let klass = line.indexOf("<class>");
			if(klass == 0){
				re = /(<class>)(.*)(<\/class>)(.*)/g
				a = re.exec(line);
				item.Klass = a[2];
				item.Text = a[4];
				msg[i] = item;
				continue;
			}
			
			msg[i] = item;
		}
		this.opieClipboard = [];
		let msgBox = document.getElementById("opieMsgBox");
		if(msgBox){
			for(let i=0; i<msg.length; i++){
				if(msg[i].Text == ""){
					msgBox.appendChild(separator.cloneNode(false));
					this.opieClipboard.push(msg[i].Text);
					continue;
				}
				let desc = document.createElement('description');
				desc.setAttribute("style",msg[i].Style);
				if(msg[i].Klass){
					desc.removeAttribute("style");
					desc.setAttribute("class",msg[i].Klass);
					if(msg[i].Klass == "link"){
						let txt = document.createElement('text');
						txt.setAttribute("value",msg[i].Text);
						txt.removeAttribute("onclick");
						clickify(txt, msg[i].Link);
						desc.appendChild(txt);
					}else{
						desc.textContent = msg[i].Text;			
					}
				}else{
					desc.textContent = msg[i].Text;			
				}
				msgBox.appendChild(desc);
				this.opieClipboard.push(msg[i].Text);
			}
			try{
				document.documentElement.getButton("cancel").focus();
			}catch(err){;;}
		}
		this.opiePlaySound('message');
		
		function clickify(elem, address) {
			elem.setAttribute("href", address);
			elem.onclick = function(event){OPIE.opieOpenLink(this.getAttribute('href'))};
			return true;
		}
		return true;
	},
	
	opieOpenLink: function (URL){
		if(!URL){return false;}
		let wm = Cc["@mozilla.org/appshell/window-mediator;1"].getService(Ci.nsIWindowMediator);
		let win = wm.getMostRecentWindow("navigator:browser");
		let currBlank = false;
		try{
			let opieGetBrowser = win.getBrowser();
			let loc = opieGetBrowser.mCurrentTab.linkedBrowser.contentDocument.location;
			if(loc == "about:blank") currBlank = true;
			if(loc == "about:newtab") currBlank = true;
			if(opieGetBrowser.mCurrentTab.label == "(Untitled)") currBlank = true;
		}catch(e){;;}
		if(currBlank){
			win.loadURI(URL);
		}else{
			win.openNewTabWith(URL, this.docURL, null, null);
		}
		return win;
	},

	opieSetAlertMsg: function (){
		this.opieSetConfirmMsg();
		return true;
	},
		
	opieSetConfirmMsg: function (){
		// If the message contains a style, it must be the first part of the message (one style per line, if needed)
		let tmp = window.arguments[0]+"   \n\n"; 
		let msg = {}; 
		let i = 0; 
		while(true){	// Break it into an array
			let item = new this.opieAlertObj(); 
			let p = tmp.indexOf("\n"); 
			if(p == -1){break;}
			item.Style = "font-family: arial; font-size: 10pt; font-weight: normal; color: purple";  // Default style
			let ps1 = tmp.indexOf("<style>"); 
			if(ps1 == 0){	
				let ps2 = tmp.indexOf("</style>"); 
				if(ps2 != -1){
					item.Style = tmp.substr(7,ps2-7);			// "7" equals the length of "<style>"
					tmp = tmp.replace(tmp.substr(0,ps2+8),"");	// Remove the style
					p = tmp.indexOf("\n");
				}
			}
			item.Text = tmp.substr(0,p);
			msg[i] = item;
			i++;
			tmp = tmp.substr(p+1);	
		}
		this.opieClipboard = [];

		let msgBox = document.getElementById("opieMsgBox"); 
		if(msgBox){
			for(let i in msg){
				let desc = document.createElement("description"); 
				desc.textContent = msg[i].Text;
				desc.setAttribute("style",msg[i].Style);
				msgBox.appendChild(desc);
				this.opieClipboard.push(msg[i].Text);
			}
		}
		try{		
			let dialog = document.documentElement; 
			let button = dialog.getButton('extra1');
			button.hidden = false;
		}catch(err){;;}
		
		let sound = OPIEstorage.getItem("alertsound", "alert"); 
		this.opiePlaySound(sound);
		
		OPIEstorage.setItem("alertsound", "alert")
		return true;
	},
	
	opieSliderLoad: function (noSlide){
		let opieMsgType = OPIE.opiePrefs.getCharPref('extensions.opie.alertType');
		
		// 'noSlide' is for resizable windows opened with window.openDialog()
		if(typeof noSlide == 'undefined') noSlide = false;
		if(noSlide) febeMsgType = "standard";
		sizeToContent();
		let finalHeight = window.outerHeight;
		let windowtype = document.documentElement.getAttribute("windowtype");

		if(opieMsgType == 'slider'){
			var sliderIncrement = 10;		// Amount the slider grows by (pixels)
			var sliderTime = 25;				// Time between increments (milliseconds)
			window.outerHeight = 1;			// First shrink the window down
			let x =(screen.availLeft + screen.availWidth - window.outerWidth) - 10;
			if(windowtype == 'opie:help') x=0;	// Help screens popup on the left
			let y = screen.availTop + screen.availHeight - window.outerHeight;
			window.moveTo(x,y);
			window.focus();
			setTimeout(function(){animateAlert();}, sliderTime);
		}
		
		function animateAlert(){
			if(window.outerHeight < finalHeight){
				window.screenY -= sliderIncrement;
				window.outerHeight += sliderIncrement;
				setTimeout(function(){animateAlert()}, sliderTime);
				return true;
			}
		}
	
		// On windows that don't have a 'Copy to clipboard' button, the button label will be blank
		let copy2clipboardBtn = document.documentElement.getButton("extra1");
		let copy2clipboard = OPIE.opiePrefs.getBoolPref("extensions.opie.copyToClipboard");
		copy2clipboardBtn.hidden = (!copy2clipboard || (copy2clipboardBtn.label == ""));
		try{
			document.documentElement.getButton("cancel").focus();
		}catch(err){;;}
		return true;
	},

	opieCopyToClipboard: function (){
		// Copy the contents of a opieAlert, opieConfirm, or opieErr box to the system clipboard
		let gClipboardHelper = Cc["@mozilla.org/widget/clipboardhelper;1"].getService(Ci.nsIClipboardHelper); 
		gClipboardHelper.copyString(this.opieClipboard.join("\n"));
		this.opiePlaySound('success');
		return true;
	},

	opieLoadHelp: function (){
		this.opieSetMsgs(); 
		let index = window.arguments[0]; 
		let field = window.arguments[1]; 
		const HTMLNS = "http://www.w3.org/1999/xhtml";

		document.getElementById("field").value = field;
		let fh = document.getElementById('opie-help'); 
		let helpTxt = ""; 
		
		// XUL pallette
		let descElm = document.createElement('description'); 
		let separatorElm = document.createElement('separator'); 
		separatorElm.setAttribute('class', "thin");
		let vboxElm = document.createElement('vbox'); 
		let hboxElm = document.createElement('hbox'); 
		let gridElm = document.createElement('grid'); 
		let columnsElm = document.createElement('columns'); 
		let columnElm = document.createElement('column'); 
		let rowsElm = document.createElement('rows'); 
		let rowElm = document.createElement('row'); 
		let spacerElm = document.createElement('spacer'); 
		spacerElm.setAttribute('flex', "1");
		let imageElm = document.createElement('image'); 
		let labelElm = document.createElement('label'); 
		let textElm = document.createElement('text'); 
		
		// HTML namespace elements
		let htmlulElm = document.createElementNS(HTMLNS, "html:ul"); 
		let htmlliElm = document.createElementNS(HTMLNS, "html:li"); 
		let helpTxtbox = document.getElementById('helptextbox'); 
		let desc, text, hbox;

		switch(index){
			case "01":
				for(var i=1; i<8; i++){ 
					helpTxt = fh.getString("opie.help_"+index+".text"+i);
					this.opieAddTxtHelpLine(helpTxtbox,helpTxt,true,"");
				}
				helpTxt = fh.getString("opie.help_"+index+".text8");
				this.opieAddTxtHelpLine(helpTxtbox,helpTxt,true,"fontitalic");
				break;
			case "02":
				for(var i=1; i<3; i++){ 
					helpTxt = fh.getString("opie.help_"+index+".text"+i);
					this.opieAddTxtHelpLine(helpTxtbox,helpTxt,true,"");
				}
				break;
			case "03": 
				helpTxt = fh.getString("opie.help_03.text1");
				this.opieAddTxtHelpLine(helpTxtbox,helpTxt,true,"");
				helpTxt = fh.getString("opie.help_03.text2");
				this.opieAddTxtHelpLine(helpTxtbox,helpTxt,true,"");
				
				desc = descElm.cloneNode(false); 
				text = textElm.cloneNode(false); 
				text.setAttribute('value', "extension.preference ");
				text.setAttribute('class', "normal");
				desc.appendChild(text);
				helpTxt = fh.getString("opie.help_03.text3");
				desc.appendChild(document.createTextNode(helpTxt));
				helpTxtbox.appendChild(desc);
							
				desc = descElm.cloneNode(false); 
				text = textElm.cloneNode(false); 
				text.setAttribute('value', "extension.preference ");
				text.setAttribute('class', "changed_valid ");
				desc.appendChild(text);
				helpTxt = fh.getString("opie.help_03.text4");
				desc.appendChild(document.createTextNode(helpTxt));
				helpTxtbox.appendChild(desc);
				
				desc = descElm.cloneNode(false); 
				text = textElm.cloneNode(false); 
				text.setAttribute('value', "extension.preference ");
				text.setAttribute('class', "changed_invalid");
				desc.appendChild(text);
				helpTxt = fh.getString("opie.help_03.text5");
				desc.appendChild(document.createTextNode(helpTxt));
				helpTxtbox.appendChild(desc);
				
				desc = descElm.cloneNode(false); 
				text = textElm.cloneNode(false); 
				text.setAttribute('value', this.opieMsg[13]+" ");
				text.setAttribute('class', "unknown");
				desc.appendChild(text);
				helpTxt = fh.getString("opie.help_03.text6");
				desc.appendChild(document.createTextNode(helpTxt));
				helpTxtbox.appendChild(desc);

				for(var i=7; i<11; i++){ 
					helpTxt = fh.getString("opie.help_"+index+".text"+i);
					this.opieAddTxtHelpLine(helpTxtbox,helpTxt,true,"");
				}

				break;
			case "04": 
				helpTxt = fh.getString("opie.help_"+index+".text1");
				this.opieAddTxtHelpLine(helpTxtbox,helpTxt,true,"");
										
				text = textElm.cloneNode(false); 
				helpTxt = "1) "+document.getElementById("opie.export.option.label2").value+":";
				this.opieAddTxtHelpLine(helpTxtbox,helpTxt,false,"fontbold");
				
				hbox = hboxElm.cloneNode(false); 
				hbox.setAttribute('flex', "0");
				hbox.setAttribute('class', "helpindent");
				
				text = textElm.cloneNode(false); 
				helpTxt = fh.getString("opie.help_"+index+".text2");
				text.setAttribute('value', helpTxt);
				text.setAttribute('class', "helpnormal");
				desc = descElm.cloneNode(false); 
				desc.appendChild(text);
				hbox.appendChild(desc);
				
				helpTxt = fh.getString("opie.help_"+index+".text3");
				desc = descElm.cloneNode(false); 
				desc.setAttribute('class', "fontitalic");
				desc.setAttribute('value', helpTxt);
				hbox.appendChild(desc);
				helpTxtbox.appendChild(hbox);
				
				text = textElm.cloneNode(false); 
				helpTxt = "2) "+document.getElementById("opie.export.option.label3").value+" ...";
				this.opieAddTxtHelpLine(helpTxtbox,helpTxt,false,"fontbold");
				helpTxt = fh.getString("opie.help_"+index+".text4");
				this.opieAddTxtHelpLine(helpTxtbox,helpTxt,true,"helpindent");
				
				helpTxt = document.getElementById("opie.export.overwrite.chkbox").value;
				this.opieAddTxtHelpLine(helpTxtbox,helpTxt,false,"fontbold");
				helpTxt = fh.getString("opie.help_"+index+".text5");
				this.opieAddTxtHelpLine(helpTxtbox,helpTxt,true,"helpindent");

				helpTxt = document.getElementById("opie.export.ignoreerror.chkbox").value;
				this.opieAddTxtHelpLine(helpTxtbox,helpTxt,false,"fontbold");
				helpTxt = fh.getString("opie.help_"+index+".text6");
				this.opieAddTxtHelpLine(helpTxtbox,helpTxt,true,"helpindent");
				
				break;
			case "06": 
				for(var i=1; i<4; i++){ 
					helpTxt = fh.getString("opie.help_"+index+".text"+i);
					this.opieAddTxtHelpLine(helpTxtbox,helpTxt,true,"");
				}
				
				hbox = hboxElm.cloneNode(false); 
				hbox.setAttribute('flex', "0");
				
				text = textElm.cloneNode(false); 
				helpTxt = fh.getString("opie.help_"+index+".text4");
				text.setAttribute('value', helpTxt);
				text.setAttribute('class', "helpnormal");
				desc = descElm.cloneNode(false); 
				desc.appendChild(text);
				hbox.appendChild(desc);
				
				helpTxt = fh.getString("opie.help_"+index+".text5");
				desc = descElm.cloneNode(false); 
				desc.setAttribute('class', "fontitalic");
				desc.setAttribute('value', helpTxt);
				hbox.appendChild(desc);
				helpTxtbox.appendChild(hbox);
				break;
			default:
				helpTxt = fh.getString("opie.help_"+index+".text1");
				this.opieAddTxtHelpLine(helpTxtbox,helpTxt,true,"");
				break;
		}
		this.opiePlaySound("help");
		return true;
	},

	opieAddTxtHelpLine: function (box,txt,separate,klass){
		let desc = document.createElement('description'); 
		desc.setAttribute('flex', "1");
		if(klass != ""){
			desc.setAttribute('class', klass);
		}else{
			desc.setAttribute('class', "helpnormal");
		}
		desc.textContent = txt;
		box.appendChild(desc);
		let separator = document.createElement('separator'); 
		if(separate){
			separator.setAttribute('class', "thin");
			box.appendChild(separator);
		}
	}
}